/*
 * Copyright (C) 2021 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package android.service.screenshot;

import static com.android.internal.util.function.pooled.PooledLambda.obtainMessage;

import android.annotation.NonNull;
import android.annotation.Nullable;
import android.annotation.SystemApi;
import android.app.Service;
import android.content.Intent;
import android.graphics.Rect;
import android.hardware.HardwareBuffer;
import android.os.Bundle;
import android.os.Handler;
import android.os.IBinder;
import android.os.Looper;
import android.os.RemoteCallback;

/**
 * A service that handles generating and verify {@link ScreenshotHash}.
 *
 * The service will generate a ScreenshotHash based on arguments passed in. Then later that
 * same ScreenshotHash can be verified to determine that it was created by the system.
 *
 * @hide
 */
@SystemApi
public abstract class ScreenshotHasherService extends Service {
    /** @hide **/
    public static final String EXTRA_SCREENSHOT_HASH =
            "android.service.screenshot.extra.SCREENSHOT_HASH";

    /** @hide **/
    public static final String EXTRA_VERIFICATION_STATUS =
            "android.service.screenshot.extra.VERIFICATION_STATUS";

    /**
     * Manifest metadata key for the resource string array containing the names of all hashing
     * algorithms provided by the service.
     *
     * @hide
     */
    public static final String SERVICE_META_DATA_KEY_AVAILABLE_ALGORITHMS =
            "android.screenshot.available_algorithms";

    /**
     * The {@link Intent} action that must be declared as handled by a service in its manifest
     * for the system to recognize it as a ScreenshotHash providing service.
     *
     * @hide
     */
    @SystemApi
    public static final String SERVICE_INTERFACE =
            "android.service.screenshot.ScreenshotHasherService";

    private ScreenshotHasherServiceWrapper mWrapper;
    private Handler mHandler;

    public ScreenshotHasherService() {
    }

    @Override
    public void onCreate() {
        super.onCreate();
        mWrapper = new ScreenshotHasherServiceWrapper();
        mHandler = new Handler(Looper.getMainLooper(), null, true);
    }

    @NonNull
    @Override
    public final IBinder onBind(@NonNull Intent intent) {
        return mWrapper;
    }

    /**
     * Generates the ScreenshotHash that can be used to validate that the system generated the
     * token.
     *
     * @param salt          The salt to use when generating the hmac. This should be unique to the
     *                      caller so the token cannot be verified by any other process.
     * @param screenshot    The screenshot buffer for the content.
     * @param bounds        The size and position of the content being screenshot in the window.
     * @param hashAlgorithm The String for the hashing algorithm to use based values in
     *                      {@link #SERVICE_META_DATA_KEY_AVAILABLE_ALGORITHMS)}.
     * @return A ScreenshotHash that can be used to validate information about the content.
     * Returns null when the arguments sent are invalid.
     */
    @Nullable
    public abstract ScreenshotHash onGenerateScreenshotHash(@NonNull byte[] salt,
            @NonNull HardwareBuffer screenshot, @NonNull Rect bounds,
            @NonNull String hashAlgorithm);

    /**
     * Call to verify that the ScreenshotHash passed in was generated by the system.
     *
     * @param salt               The salt value to use when verifying the hmac. This should be the
     *                           same value that was passed to
     *                           {@link #onGenerateScreenshotHash(byte[],
     *                           HardwareBuffer, Rect, String)} to
     *                           generate the token.
     * @param screenshotHash The token to verify that it was generated by the system.
     * @return true if the token can be verified that it was generated by the system.
     */
    public abstract boolean onVerifyScreenshotHash(@NonNull byte[] salt,
            @NonNull ScreenshotHash screenshotHash);

    private void generateScreenshotHash(byte[] salt, HardwareBuffer screenshot, Rect bounds,
            String hashAlgorithm, RemoteCallback callback) {
        ScreenshotHash screenshotHash = onGenerateScreenshotHash(salt, screenshot,
                bounds,
                hashAlgorithm);
        final Bundle data = new Bundle();
        data.putParcelable(EXTRA_SCREENSHOT_HASH, screenshotHash);
        callback.sendResult(data);
    }

    private void verifyScreenshotHash(byte[] salt, ScreenshotHash screenshotHash,
            RemoteCallback callback) {
        boolean verificationStatus = onVerifyScreenshotHash(salt, screenshotHash);
        final Bundle data = new Bundle();
        data.putBoolean(EXTRA_VERIFICATION_STATUS, verificationStatus);
        callback.sendResult(data);
    }

    private final class ScreenshotHasherServiceWrapper extends
            IScreenshotHasherService.Stub {
        @Override
        public void generateScreenshotHash(byte[] salt, HardwareBuffer screenshot, Rect bounds,
                String hashAlgorithm, RemoteCallback callback) {
            mHandler.sendMessage(
                    obtainMessage(ScreenshotHasherService::generateScreenshotHash,
                            ScreenshotHasherService.this, salt, screenshot, bounds,
                            hashAlgorithm, callback));
        }

        @Override
        public void verifyScreenshotHash(byte[] salt, ScreenshotHash screenshotHash,
                RemoteCallback callback) {
            mHandler.sendMessage(
                    obtainMessage(ScreenshotHasherService::verifyScreenshotHash,
                            ScreenshotHasherService.this, salt, screenshotHash,
                            callback));
        }
    }
}
